import re
import unicodedata
from difflib import SequenceMatcher

import requests

# from reversion import revisions as reversion
import reversion
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.query import QuerySet
from django.urls import reverse
from django.utils.functional import cached_property
from math import inf
from model_utils.managers import InheritanceManager
import copy

from apis_core.apis_entities.serializers_generic import EntitySerializer
from apis_core.apis_labels.models import Label
from apis_core.apis_vocabularies.models import CollectionType, LabelType, TextType

from django.contrib.contenttypes.fields import GenericRelation
# from helper_functions.highlighter import highlight_text
from apis_core.default_settings.NER_settings import autocomp_settings
from apis_core.helper_functions import DateParser

NEXT_PREV = getattr(settings, "APIS_NEXT_PREV", True)


@reversion.register()
class TempEntityClass(models.Model):
    """Base class to bind common attributes to many classes.

    The common attributes are:
    written start and enddates
    recognized start and enddates which are derived by RegEx
    from the written dates.
    A review boolean field to mark an object as reviewed
    """

    name = models.CharField(max_length=255, blank=True)
    review = models.BooleanField(
        default=False,
        help_text="Should be set to True, if the data record holds up quality standards.",
    )
    start_date = models.DateField(blank=True, null=True)
    start_start_date = models.DateField(blank=True, null=True)
    start_end_date = models.DateField(blank=True, null=True)
    end_date = models.DateField(blank=True, null=True)
    end_start_date = models.DateField(blank=True, null=True)
    end_end_date = models.DateField(blank=True, null=True)
    start_date_written = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name="Start",
    )
    end_date_written = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name="End",
    )
    text = models.ManyToManyField("Text", blank=True)
    collection = models.ManyToManyField("Collection")
    status = models.CharField(max_length=100)
    source = models.ForeignKey(
        "Source", blank=True, null=True, on_delete=models.SET_NULL
    )
    references = models.TextField(blank=True, null=True)
    notes = models.TextField(blank=True, null=True)
    published = models.BooleanField(default=False)
    objects = models.Manager()
    objects_inheritance = InheritanceManager()

    if "apis_highlighter" in settings.INSTALLED_APPS:
        from apis_highlighter.models import Annotation
        annotation_set = GenericRelation(Annotation)

    def __str__(self):
        if self.name != "" and hasattr(
            self, "first_name"
        ):  # relation usually don´t have names
            return "{}, {} (ID: {})".format(self.name, self.first_name, self.id)
        elif self.name != "":
            return "{} (ID: {})".format(self.name, self.id)
        else:
            return "(ID: {})".format(self.id)

    def save(self, parse_dates=True, *args, **kwargs):
        """Adaption of the save() method of the class to automatically parse string-dates into date objects"""

        if parse_dates:

            # overwrite every field with None as default
            start_date = None
            start_start_date = None
            start_end_date = None
            end_date = None
            end_start_date = None
            end_end_date = None

            if self.start_date_written:
                # If some textual user input of a start date is there, then parse it

                start_date, start_start_date, start_end_date = DateParser.parse_date(
                    self.start_date_written
                )

            if self.end_date_written:
                # If some textual user input of an end date is there, then parse it

                end_date, end_start_date, end_end_date = DateParser.parse_date(
                    self.end_date_written
                )

            self.start_date = start_date
            self.start_start_date = start_start_date
            self.start_end_date = start_end_date
            self.end_date = end_date
            self.end_start_date = end_start_date
            self.end_end_date = end_end_date

        if self.name:
            self.name = unicodedata.normalize("NFC", self.name)

        super(TempEntityClass, self).save(*args, **kwargs)

        return self

    def get_child_entity(self):
        for x in [x for x in apps.all_models["apis_entities"].values()]:
            if x.__name__ in list(settings.APIS_ENTITIES.keys()):
                try:
                    my_ent = x.objects.get(id=self.id)
                    return my_ent
                    break
                except ObjectDoesNotExist:
                    pass
        return None

    @classmethod
    def get_listview_url(self):
        entity = self.__name__.lower()
        if entity == "institution" or len(entity) < 10:
            return reverse(
                "apis_core:apis_entities:generic_entities_list",
                kwargs={"entity": entity},
            )
        else:
            return reverse(
                "apis_core:apis_relations:generic_relations_list",
                kwargs={"entity": entity},
            )

    @classmethod
    def get_createview_url(self):
        entity = self.__name__.lower()
        if entity == "institution" or len(entity) < 10:
            return reverse(
                "apis_core:apis_entities:generic_entities_create_view",
                kwargs={"entity": entity},
            )
        else:
            return None

    def get_edit_url(self):
        entity = self.__class__.__name__.lower()
        if entity == "institution" or len(entity) < 10:
            return reverse(
                "apis_core:apis_entities:generic_entities_edit_view",
                kwargs={"entity": entity, "pk": self.id},
            )
        else:
            return None

    def get_child_class(self):
        child = self.get_child_entity()
        if child:
            return "{}".format(child.__class__.__name__)
        else:
            return "{}".format(child.__class__.__name__)

    def get_absolute_url(self):
        entity = self.__class__.__name__.lower()
        if entity == "institution" or len(entity) < 10:
            return reverse(
                "apis_core:apis_entities:generic_entities_detail_view",
                kwargs={"entity": entity, "pk": self.id},
            )
        elif entity == "tempentityclass":
            return self.get_child_entity().get_absolute_url()
        else:
            return reverse(
                "apis_core:apis_relations:generic_relations_detail_view",
                kwargs={"entity": entity, "pk": self.id},
            )

    def get_prev_url(self):
        entity = self.__class__.__name__.lower()
        if NEXT_PREV:
            prev = self.__class__.objects.filter(id__lt=self.id).order_by("-id")
        else:
            return False
        if prev:
            if entity == "institution" or len(entity) < 10:
                return reverse(
                    "apis_core:apis_entities:generic_entities_detail_view",
                    kwargs={"entity": entity, "pk": prev.first().id},
                )
            else:
                return reverse(
                    "apis_core:apis_relations:generic_relations_detail_view",
                    kwargs={"entity": entity, "pk": prev.first().id},
                )
        else:
            return False

    def get_next_url(self):
        entity = self.__class__.__name__.lower()
        if NEXT_PREV:
            next = self.__class__.objects.filter(id__gt=self.id)
        else:
            return False
        if next:
            if entity == "institution" or len(entity) < 10:
                return reverse(
                    "apis_core:apis_entities:generic_entities_detail_view",
                    kwargs={"entity": entity, "pk": next.first().id},
                )
            else:
                return reverse(
                    "apis_core:apis_relations:generic_relations_detail_view",
                    kwargs={"entity": entity, "pk": next.first().id},
                )
        else:
            return False

    def get_delete_url(self):
        entity = self.__class__.__name__.lower()
        if entity == "institution" or len(entity) < 10:
            return reverse(
                "apis_core:apis_entities:generic_entities_delete_view",
                kwargs={"entity": entity, "pk": self.id},
            )
        else:
            return None

    def merge_with(self, entities):
        e_a_pk = self.pk
        e_a = type(self).__name__
        self_model_class = ContentType.objects.get(
            app_label="apis_entities", model__iexact=e_a
        ).model_class()
        if isinstance(entities, int):
            entities = self_model_class.objects.get(pk=entities)
        if not isinstance(entities, list) and not isinstance(entities, QuerySet):
            entities = [entities]
            entities = [
                self_model_class.objects.get(pk=ent) if type(ent) == int else ent
                for ent in entities
            ]
        rels = ContentType.objects.filter(
            app_label="apis_relations", model__icontains=e_a
        )
        print(entities)
        for ent in entities:
            e_b = type(ent).__name__
            e_b_pk = ent.pk
            if e_b_pk == e_a_pk:
                raise ValueError("You can not merge an entity with itself")
            if e_a != e_b:
                continue
            lt, created = LabelType.objects.get_or_create(name="Legacy name (merge)")
            col_list = list(self.collection.all())
            for col2 in ent.collection.all():
                if col2 not in col_list:
                    self.collection.add(col2)
            for f in ent._meta.local_many_to_many:
                if not f.name.endswith("_set"):
                    sl = list(getattr(self, f.name).all())
                    for s in getattr(ent, f.name).all():
                        if s not in sl:
                            getattr(self, f.name).add(s)
            Label.objects.create(label=str(ent), label_type=lt, temp_entity=self)
            for u in Uri.objects.filter(entity=ent):
                u.entity = self
                u.save()
            for l in Label.objects.filter(temp_entity=ent):
                l.temp_entity = self
                l.save()
            for r in rels.filter(model__icontains=e_b):
                lst_ents_rel = str(r).split()
                if lst_ents_rel[-1] == lst_ents_rel[-2]:
                    q_d = {"related_{}A".format(e_b.lower()): ent}
                    k = r.model_class().objects.filter(**q_d)
                    for t in k:
                        setattr(t, "related_{}A".format(e_a.lower()), self)
                        t.save()
                    q_d = {"related_{}B".format(e_b.lower()): ent}
                    k = r.model_class().objects.filter(**q_d)
                    for t in k:
                        setattr(t, "related_{}B".format(e_a.lower()), self)
                        t.save()
                else:
                    q_d = {"related_{}".format(e_b.lower()): ent}
                    k = r.model_class().objects.filter(**q_d)
                    for t in k:
                        setattr(t, "related_{}".format(e_a.lower()), self)
                        t.save()

            ent.delete()

    def get_serialization(self):
        return EntitySerializer(self).data


@reversion.register()
class Source(models.Model):
    """ Holds information about entities and their relations"""

    orig_filename = models.CharField(max_length=255, blank=True)
    indexed = models.BooleanField(default=False)
    pubinfo = models.CharField(max_length=400, blank=True)
    author = models.CharField(max_length=255, blank=True)
    orig_id = models.PositiveIntegerField(blank=True, null=True)

    def __str__(self):
        if self.author != "" and self.orig_filename != "":
            return "{}, stored by {}".format(self.orig_filename, self.author)
        else:
            return "(ID: {})".format(self.id)


@reversion.register()
class Collection(models.Model):
    """ Allows to group entities and relation. """

    name = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    collection_type = models.ForeignKey(
        CollectionType, blank=True, null=True, on_delete=models.SET_NULL
    )
    groups_allowed = models.ManyToManyField(Group)
    parent_class = models.ForeignKey(
        "self", blank=True, null=True, on_delete=models.CASCADE
    )
    published = models.BooleanField(default=False)

    @classmethod
    def from_db(cls, db, field_names, values):
        instance = super().from_db(db, field_names, values)
        instance._loaded_values = dict(zip(field_names, values))
        return instance

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if hasattr(self, "_loaded_values"):
            if self.published != self._loaded_values["published"]:
                for ent in self.tempentityclass_set.all():
                    ent.published = self.published
                    ent.save()
        super().save(*args, **kwargs)


@reversion.register()
class Text(models.Model):
    """Holds unstructured text associeted with
    one ore many entities/relations."""

    kind = models.ForeignKey(TextType, blank=True, null=True, on_delete=models.SET_NULL)
    text = models.TextField(blank=True)
    source = models.ForeignKey(Source, blank=True, null=True, on_delete=models.SET_NULL)

    def __str__(self):
        if self.text != "":
            return "ID: {} - {}".format(self.id, self.text[:25])
        else:
            return "ID: {}".format(self.id)

    def check_for_deleted_annotations(self):

        from apis_highlighter.models import Annotation
        if self.pk is not None:
            deleted = []
            orig = Text.objects.get(pk=self.pk)
            if orig.text != self.text and "apis_highlighter" in settings.INSTALLED_APPS:
                ann = Annotation.objects.filter(text_id=self.pk).order_by("start")
                min_ann_len = min([x.end-x.start for x in ann])
                seq = SequenceMatcher(lambda x: len(x) > min_ann_len, orig.text, self.text)
                for a in ann:
                    changed = False
                    count = 0
                    for s in seq.get_matching_blocks():
                        count += 1
                        if s.a <= a.start and (s.a + s.size) >= a.end:
                            old_start = copy.deepcopy(a.start)
                            old_end = copy.deepcopy(a.end)
                            new_start = a.start + (s.b - s.a)
                            new_end =  a.end + (s.b - s.a)
                            if orig.text[old_start:old_end] == self.text[new_start:new_end]:
                                changed = True
                                break
                    if not changed:
                        deleted.append(a.id)
        else:
            deleted = None
        return deleted

    def save(self, *args, **kwargs):

        if self.pk is not None:
            orig = Text.objects.get(pk=self.pk)
            if orig.text != self.text and "apis_highlighter" in settings.INSTALLED_APPS:
                from apis_highlighter.models import Annotation
                def calculate_context_weights(text, i_start, i_end):

                    def calculate_word_dict(text, direction):

                        word_list = re.split(" |\\n", text)
                        word_list = [w for w in word_list if w != ""]

                        word_dict = {}

                        if word_list != []:

                            value_step = 1 / len(word_list)
                            value_current = 1
                            for word in word_list[::direction]:
                                word_value = word_dict.get(word, 0)
                                word_dict[word] = word_value + value_current
                                value_current -= value_step

                        return word_dict

                    text_left = text[:i_start]
                    text_right = text[i_end:]
                    word_dict_left = calculate_word_dict(text=text_left, direction=-1)
                    word_dict_right = calculate_word_dict(text=text_right, direction=1)

                    return {
                        "word_dict_left": word_dict_left,
                        "word_dict_right": word_dict_right,
                    }

                def make_diff(word_dict_a, word_dict_b):

                    words_all = set(word_dict_a.keys()).union(set(word_dict_b.keys()))
                    diff_all = 0
                    for word in words_all:
                        word_value_a = word_dict_a.get(word, 0)
                        word_value_b = word_dict_b.get(word, 0)
                        diff_all += abs(word_value_a - word_value_b)

                    return diff_all

                def correlate_annotations(text_old, text_new, annotations_old):

                    for ann in annotations_old:
                        i_old_start = ann.start
                        i_old_end = ann.end
                        context_weights_dict_old = calculate_context_weights(
                            text=text_old,
                            i_start=i_old_start,
                            i_end=i_old_end
                        )
                        ann_text = text_old[ann.start:ann.end]
                        diff_min = inf
                        i_new_start = None
                        i_new_end = None
                        for i in re.finditer(f"(?={re.escape(ann_text)})", text_new):
                            i_candidate_start = i.start()
                            i_candidate_end = i_candidate_start + len(ann_text)
                            context_weights_dict_new = calculate_context_weights(
                                text=text_new, i_start=i_candidate_start, i_end=i_candidate_end
                            )
                            diff_left = make_diff(
                                context_weights_dict_new["word_dict_left"],
                                context_weights_dict_old["word_dict_left"]
                            )
                            diff_right = make_diff(
                                context_weights_dict_new["word_dict_right"],
                                context_weights_dict_old["word_dict_right"]
                            )
                            diff_current = diff_left + diff_right
                            if diff_current < diff_min:
                                diff_min = diff_current
                                i_new_start = i_candidate_start
                                i_new_end = i_candidate_end

                        if diff_min != inf:
                            ann.start = i_new_start
                            ann.end = i_new_end
                            ann.save()
                        else:
                            ann.delete() # TODO: we might want to delete relations as well.

                correlate_annotations(
                    text_old=orig.text,
                    text_new=self.text,
                    annotations_old=Annotation.objects.filter(text_id=self.pk).order_by("start")
                )

        super().save(*args, **kwargs)


@reversion.register()
class Uri(models.Model):
    uri = models.URLField(blank=True, null=True, unique=True, max_length=255)
    domain = models.CharField(max_length=255, blank=True)
    rdf_link = models.URLField(blank=True)
    entity = models.ForeignKey(
        TempEntityClass, blank=True, null=True, on_delete=models.CASCADE
    )
    # loaded set to True when RDF was loaded and parsed into the data model
    loaded = models.BooleanField(default=False)
    # Timestamp when file was loaded and parsed
    loaded_time = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return str(self.uri)

    def get_web_object(self):
        result = {
            "relation_pk": self.pk,
            "relation_type": "uri",
            "related_entity": self.entity.name,
            "related_entity_url": self.entity.get_absolute_url(),
            "related_entity_class_name": self.entity.__class__.__name__.lower(),
            "uri": self.uri,
        }
        return result

    @classmethod
    def get_listview_url(self):
        return reverse("apis_core:apis_metainfo:uri_browse")

    @classmethod
    def get_createview_url(self):
        return reverse("apis_core:apis_metainfo:uri_create")

    def get_absolute_url(self):
        return reverse("apis_core:apis_metainfo:uri_detail", kwargs={"pk": self.id})

    def get_delete_url(self):
        return reverse("apis_core:apis_metainfo:uri_delete", kwargs={"pk": self.id})

    def get_edit_url(self):
        return reverse("apis_core:apis_metainfo:uri_edit", kwargs={"pk": self.id})


@reversion.register()
class UriCandidate(models.Model):
    """Used to store the URI candidates for automatically generated entities."""

    uri = models.URLField()
    confidence = models.FloatField(blank=True, null=True)
    responsible = models.CharField(max_length=255)
    entity = models.ForeignKey(
        TempEntityClass, blank=True, null=True, on_delete=models.CASCADE
    )

    @cached_property
    def description(self):
        headers = {"accept": "application/json"}
        cn = TempEntityClass.objects_inheritance.get_subclass(
            id=self.entity_id
        ).__class__.__name__
        for endp in autocomp_settings[cn.title()]:
            url = re.sub(r"/[a-z]+$", "/entity", endp["url"])
            params = {"id": self.uri}
            res = requests.get(url, params=params, headers=headers)
            if res.status_code == 200:
                if endp["fields"]["descr"][0] in res.json()["representation"].keys():
                    desc = res.json()["representation"][endp["fields"]["descr"][0]][0][
                        "value"
                    ]
                else:
                    desc = "undefined"
                label = res.json()["representation"][endp["fields"]["name"][0]][0][
                    "value"
                ]
                return (label, desc)


# @receiver(post_save, sender=Uri, dispatch_uid="remove_default_uri")
# def remove_default_uri(sender, instance, **kwargs):
#    if Uri.objects.filter(entity=instance.entity).count() > 1:
#        Uri.objects.filter(entity=instance.entity, domain="apis default").delete()



from django.conf import settings


def access_for_all(self, viewtype="list"):
    if self.request.user.is_authenticated:
        return self.request.user.is_authenticated
    elif viewtype == "list":
        try:
            access = settings.APIS_LIST_VIEWS_ALLOWED
        except AttributeError:
            access = False
            return access
        return access
    else:
        try:
            access = settings.APIS_DETAIL_VIEWS_ALLOWED
        except AttributeError:
            access = False
            return access
        return access


def access_for_all_function(user):
    if user.is_anonymous:
        print(getattr(settings, 'APIS_DETAIL_VIEWS_ALLOWED', False))
        return getattr(settings, 'APIS_DETAIL_VIEWS_ALLOWED', False)
    else:
        return True


ENTITIES_DEFAULT_COLS = [
    'start_date',
    'start_date_written',
    'end_date',
    'end_date_written',
    'text',
    'collection',
    'status',
    'source',
    'references',
    'notes',
]

def get_child_classes(objids, obclass, labels=False):
    """used to retrieve a list of primary keys of sub classes"""
    if labels:
        labels_lst = []
    for obj in objids:
        obj = obclass.objects.get(pk=obj)
        p_class = list(obj.vocabsbaseclass_set.all())
        p = p_class.pop() if len(p_class) > 0 else False
        while p:
            if p.pk not in objids:
                if labels:
                    labels_lst.append((p.pk, p.label))
                objids.append(p.pk)
            p_class += list(p.vocabsbaseclass_set.all())
            p = p_class.pop() if len(p_class) > 0 else False
    if labels:
        return (objids, labels_lst)
    else:
        return objids



import json
import re
import inspect

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponse, Http404
from django.template.loader import render_to_string
from apis_core.apis_relations import forms as relation_form_module

from apis_core.apis_entities.models import Person, Institution, Place, Event, Work, AbstractEntity
from apis_core.apis_labels.models import Label
from apis_core.apis_metainfo.models import Uri
from .forms2 import GenericRelationForm
from .models import (
    PersonPlace, PersonPerson, PersonInstitution, InstitutionPlace,
    InstitutionInstitution, PlacePlace, PersonEvent, InstitutionEvent, PlaceEvent, PersonWork,
    InstitutionWork, PlaceWork, EventWork, WorkWork
)
#from .forms import PersonLabelForm, InstitutionLabelForm, PlaceLabelForm, EventLabelForm
from .tables import LabelTableEdit

form_module_list = [relation_form_module]

if 'apis_highlighter' in settings.INSTALLED_APPS:
    from apis_highlighter.highlighter import highlight_text_new
    from apis_highlighter import forms as highlighter_form_module
    form_module_list.append(highlighter_form_module)


def turn_form_modules_into_dict(form_module_list):
    """
    Since form classes are loaded dynamically from the respective modules and it's settings-dependent which modules
    are imported and which not, it's better to differentiate here which modules are imported close to their imports
    and then providing a dict for later extraction of the required form class.
    """

    form_class_dict = {}
    for m in form_module_list:
        for name, cls in inspect.getmembers(m, inspect.isclass):
            form_class_dict[name] = cls

    return form_class_dict

form_class_dict = turn_form_modules_into_dict(form_module_list)


############################################################################
############################################################################
#
#   Generic views for AjaxForms
#
############################################################################
############################################################################

######################################################
# test for class-ignoring _ajax_form-functions
######################################################


# Model-classes must be registered together with their ModelForm-classes
registered_forms = {'WorkWorkForm': [WorkWork, Work, Work],
                    'PersonPlaceForm': [PersonPlace, Person, Place],
                    'PersonPlaceHighlighterForm': [PersonPlace, Person, Place],
                    'PersonPersonForm': [PersonPerson, Person, Person],
                    'PersonPersonHighlighterForm': [PersonPerson, Person, Person],
                    'PersonInstitutionForm': [PersonInstitution, Person, Institution],
                    'PersonEventForm': [PersonEvent, Person, Event],
                    'PersonWorkForm': [PersonWork, Person, Work],
                    'PersonInstitutionHighlighterForm': [PersonInstitution, Person, Institution],
                    'PersonWorkHighlighterForm': [PersonWork, Person, Work],
                    'PlaceWorkHighlighterForm': [PlaceWork, Place, Work],
                    'InstitutionWorkHighlighterForm': [InstitutionWork, Institution, Work],
                    'InstitutionPlaceForm': [InstitutionPlace, Institution, Place],
                    'InstitutionInstitutionForm': [
                        InstitutionInstitution,
                        Institution,
                        Institution],
                    'InstitutionPersonForm': [PersonInstitution, Institution, Person],
                    'InstitutionEventForm': [InstitutionEvent, Institution, Event],
                    'InstitutionWorkForm': [InstitutionWork, Institution, Work],
                    'PlaceEventForm': [PlaceEvent, Place, Event],
                    'PlaceWorkForm': [PlaceWork, Place, Work],
                    'PlacePlaceForm': [PlacePlace, Place, Place],
                    'EventWorkForm': [EventWork, Event, Work],
                    'InstitutionLabelForm': [Label, Institution, Label],
                    'PersonLabelForm': [Label, Person, Label],
                    'EventLabelForm': [Label, Event, Label],
                    'PersonResolveUriForm': [Uri, Person, Uri],
                    'SundayHighlighterForm': [ ],
                    'AddRelationHighlighterPersonForm': [],
                    #'PlaceHighlighterForm': [Annotation, ],
                    #'PersonHighlighterForm': [Annotation, ]
                    }


@login_required
def get_form_ajax(request):
    '''Returns forms rendered in html'''

    FormName = request.POST.get('FormName')
    SiteID = request.POST.get('SiteID')
    ButtonText = request.POST.get('ButtonText')
    ObjectID = request.POST.get('ObjectID')
    entity_type_str = request.POST.get('entity_type')
    form_match = re.match(r'([A-Z][a-z]+)([A-Z][a-z]+)(Highlighter)?Form', FormName)
    form_match2 = re.match(r'([A-Z][a-z]+)(Highlighter)?Form', FormName)
    if FormName and form_match:
        entity_type_v1 = ContentType.objects.filter(
            model='{}{}'.format(form_match.group(1).lower(), form_match.group(2)).lower(),
            app_label='apis_relations')
        entity_type_v2 = ContentType.objects.none()
    elif FormName and form_match2:
        entity_type_v2 = ContentType.objects.filter(
            model='{}'.format(
                form_match.group(1).lower(),
                app_label='apis_entities'))
        entity_type_v1 = ContentType.objects.none()
    else:
        entity_type_v1 = ContentType.objects.none()
        entity_type_v2 = ContentType.objects.none()
    if ObjectID == 'false' or ObjectID is None or ObjectID == 'None':
        ObjectID = False
        form_dict = {'entity_type': entity_type_str}
    elif entity_type_v1.count() > 0:
        d = entity_type_v1[0].model_class().objects.get(pk=ObjectID)
        form_dict = {'instance': d, 'siteID': SiteID, 'entity_type': entity_type_str}
    elif entity_type_v2.count() > 0:
        d = entity_type_v2[0].model_class().objects.get(pk=ObjectID)
        form_dict = {'instance': d, 'siteID': SiteID, 'entity_type': entity_type_str}
    else:
        if FormName not in registered_forms.keys():
            raise Http404
        d = registered_forms[FormName][0].objects.get(pk=ObjectID)
        form_dict = {'instance': d, 'siteID': SiteID, 'entity_type': entity_type_str}
    if entity_type_v1.count() > 0:
        form_dict['relation_form'] = '{}{}'.format(form_match.group(1), form_match.group(2))
        if form_match.group(3) == 'Highlighter':
            form_dict['highlighter'] = True
        form = GenericRelationForm(**form_dict)
    else:
        form_class = form_class_dict[FormName]
        form = form_class(**form_dict)
    tab = FormName[:-4]
    data = {'tab': tab, 'form': render_to_string("apis_relations/_ajax_form.html", {
                "entity_type": entity_type_str,
                "form": form,
                'type1': FormName,
                'url2': 'save_ajax_'+FormName,
                'button_text': ButtonText,
                'ObjectID': ObjectID,
                'SiteID': SiteID})}

    return HttpResponse(json.dumps(data), content_type='application/json')


@login_required
def save_ajax_form(request, entity_type, kind_form, SiteID, ObjectID=False):
    '''Tests validity and saves AjaxForms, returns them when validity test fails'''
    if kind_form not in registered_forms.keys():
        raise Http404

    button_text = "create/modify"

    if not ObjectID:
        instance_id = ''
    else:
        instance_id = ObjectID
    entity_type_str = entity_type
    entity_type = AbstractEntity.get_entity_class_of_name(entity_type)

    form_match = re.match(r'([A-Z][a-z]+)([A-Z][a-z]+)?(Highlighter)?Form', kind_form)
    form_dict = {'data': request.POST,
                 'entity_type': entity_type,
                 'request': request}

    test_form_relations = ContentType.objects.filter(
        model='{}{}'.format(form_match.group(1).lower(), form_match.group(2)).lower(),
        app_label='apis_relations')
    tab = re.match(r'(.*)Form', kind_form).group(1)
    call_function = 'EntityRelationForm_response'
    if test_form_relations.count() > 0:
        relation_form = test_form_relations[0].model_class()
        form_dict['relation_form'] = relation_form
        if form_match.group(3) == 'Highlighter':
            form_dict['highlighter'] = True
            tab = form_match.group(1)+form_match.group(2)
            call_function = 'HighlForm_response'
        form = GenericRelationForm(**form_dict)
    else:
        form_class = form_class_dict[kind_form]
        form = form_class(**form_dict)
    if form.is_valid():
        site_instance = entity_type.objects.get(pk=SiteID)
        set_ann_proj = request.session.get('annotation_project', 1)
        entity_types_highlighter = request.session.get('entity_types_highlighter')
        users_show = request.session.get('users_show_highlighter', None)
        hl_text = None
        if ObjectID:
            instance = form.save(instance=ObjectID, site_instance=site_instance)
        else:
            instance = form.save(site_instance=site_instance)
        right_card = True
        if test_form_relations.count() > 0:
            table_html = form.get_html_table(entity_type_str, request, site_instance, form_match)
        if 'Highlighter' in tab or form_match.group(3) == 'Highlighter':
            hl_text = {
                'text': highlight_text_new(form.get_text_id(),
                                       users_show=users_show,
                                       set_ann_proj=set_ann_proj,
                                       types=entity_types_highlighter)[0].strip(),
                'id': form.get_text_id()}
        if tab == 'PersonLabel':
            table_html = LabelTableEdit(
                    data=site_instance.label_set.all(),
                    prefix='PL-')
        elif tab == 'InstitutionLabel':
            table_html = LabelTableEdit(
                    data=site_instance.label_set.all(),
                    prefix='IL-')
        elif tab == 'PersonResolveUri':
            table_html = EntityUriTable(
                Uri.objects.filter(entity=site_instance),
                prefix='PURI-'
            )

        elif tab == 'AddRelationHighlighterPerson' or tab == 'PlaceHighlighter' or tab == 'PersonHighlighter' or tab == 'SundayHighlighter':
            table_html = None
            right_card = False
            call_function = 'PAddRelation_response'
            instance = None
        if instance:
            instance2 = instance.get_web_object()
        else:
            instance2 = None
        if table_html:
            table_html2 = table_html.as_html(request)
        else:
            table_html2 = None
        data = {'test': True, 'tab': tab, 'call_function': call_function,
                'instance': instance2,
                'table_html': table_html2,
                'text': hl_text,
                'right_card': right_card}
    else:
        if 'Highlighter' in tab:
            call_function = 'HighlForm_response'
        data = {'test': False, 'call_function': call_function,
                'DivID': 'div_'+kind_form+instance_id,
                'form': render_to_string("apis_relations/_ajax_form.html", context={
                    "entity_type": entity_type_str,
                    "form": form, 'type1': kind_form, 'url2': 'save_ajax_'+kind_form,
                    'button_text': button_text, 'ObjectID': ObjectID, 'SiteID': SiteID},
                    request=request)}

    # except Exception as e:
    #     print('Error in save method')
    #     print(e)
    #     data = {'test': False, 'error': json.dumps(str(e))}
    return HttpResponse(json.dumps(data), content_type='application/json')



from django.conf import settings
from django.conf.urls import url
from django.urls import path

from . import api_views

app_name = 'apis_entities'

urlpatterns = [
    url(r'^savenetworkfiles/$', api_views.SaveNetworkFiles.as_view()),
    url(r'^getorcreateentity/$', api_views.GetOrCreateEntity.as_view(), name='GetOrCreateEntity'),
    path(r'entity/<int:pk>/', api_views.GetEntityGeneric.as_view(), name="GetEntityGeneric"),
    path(r'uri/', api_views.uri_resolver, name="UriResolver"),
    path(r'getrelatedplaces/', api_views.GetRelatedPlaces.as_view(), name="GetRelatedPlaces"),
    path(r'lifepath/<int:pk>/', api_views.LifePathViewset.as_view(), name="Lifepathviewset")
]

if 'deep learning' in getattr(settings, "APIS_COMPONENTS", []) and 'apis_highlighter' in settings.INSTALLED_APPS:
    from apis_highlighter.api_views import TestDLModel
    urlpatterns.append(url(r'^nlp_model/$', TestDLModel.as_view(), name='TestDLModel'),)

if 'apis_highlighter' in settings.INSTALLED_APPS:
    from apis_highlighter.api_views import AnnotatorAgreementView, ShowOverlappingHighlights
    urlpatterns.extend([url(r'^annotatoragreement/$',
                            AnnotatorAgreementView.as_view(),
                            name='AnnotatorAgreementView'),
                       url(r'^overlappinghighlights/$',
                           ShowOverlappingHighlights.as_view(),
                           name='ShowOverlappingHighlights')])



# -*- coding: utf-8 -*-
import json

import reversion
from django.conf import settings
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.shortcuts import render, redirect, get_object_or_404
from django.urls import reverse
from django.urls import reverse_lazy
from django_tables2 import RequestConfig
from django_tables2 import SingleTableView
from django_tables2.export.views import ExportMixin

# from reversion_compare.views import HistoryCompareDetailView

from apis_core.apis_metainfo.models import Uri, UriCandidate, Text
from apis_core.apis_relations.models import AbstractRelation
from apis_core.helper_functions.RDFParser import RDFParser
from apis_core.helper_functions.stanbolQueries import retrieve_obj
from apis_core.helper_functions.utils import (
    access_for_all,
    access_for_all_function,
    ENTITIES_DEFAULT_COLS,
)
from .filters import get_list_filter_of_entity
from .forms import (
    GenericFilterFormHelper,
    NetworkVizFilterForm,
    PersonResolveUriForm,
    GenericEntitiesStanbolForm,
)
from .models import Place
from .tables import get_entities_table

if "apis_highlighter" in settings.INSTALLED_APPS:
    from apis_highlighter.forms import SelectAnnotationProject
    from apis_highlighter.highlighter import highlight_text_new

if "charts" in settings.INSTALLED_APPS:
    from charts.models import ChartConfig
    from charts.views import create_payload

###########################################################################
############################################################################
#
#   Helper Functions
#
############################################################################
############################################################################


@user_passes_test(access_for_all_function)
def set_session_variables(request):
    ann_proj_pk = request.GET.get("project", None)
    types = request.GET.getlist("types", None)
    users_show = request.GET.getlist("users_show", None)
    edit_views = request.GET.get("edit_views", False)
    if types:
        request.session["entity_types_highlighter"] = types
    if users_show:
        request.session["users_show_highlighter"] = users_show
    if ann_proj_pk:
        request.session["annotation_project"] = ann_proj_pk
    if edit_views:
        if edit_views != "false":
            request.session["edit_views"] = True
    return request


@user_passes_test(access_for_all_function)
def get_highlighted_texts(request, instance):
    if "apis_highlighter" in settings.INSTALLED_APPS:
        set_ann_proj = request.session.get("annotation_project", 1)
        entity_types_highlighter = request.session.get("entity_types_highlighter", None)
        users_show = request.session.get("users_show_highlighter", None)
        object_texts = [
            {
                "text": highlight_text_new(
                    x,
                    set_ann_proj=set_ann_proj,
                    types=entity_types_highlighter,
                    users_show=users_show,
                )[0].strip(),
                "id": x.pk,
                "kind": x.kind,
            }
            for x in Text.objects.filter(tempentityclass=instance)
        ]
        ann_proj_form = SelectAnnotationProject(
            set_ann_proj=set_ann_proj,
            entity_types_highlighter=entity_types_highlighter,
            users_show_highlighter=users_show,
        )
        return object_texts, ann_proj_form
    else:
        object_texts = [
            {"text": x.text, "id": x.pk, "kind": x.kind}
            for x in Text.objects.filter(tempentityclass=instance)
        ]
        return object_texts, False


############################################################################
############################################################################
#
#   GenericViews
#
############################################################################
############################################################################


class GenericListViewNew(UserPassesTestMixin, ExportMixin, SingleTableView):
    formhelper_class = GenericFilterFormHelper
    context_filter_name = "filter"
    paginate_by = 25
    template_name = getattr(
        settings, "APIS_LIST_VIEW_TEMPLATE", "apis:apis_entities/generic_list.html"
    )
    #login_url = "/accounts/login/"

    def get_model(self):
        model = ContentType.objects.get(
            app_label__startswith="apis_", model=self.entity.lower()
        ).model_class()
        return model

    def test_func(self):
        access = access_for_all(self, viewtype="list")
        if access:
            self.request = set_session_variables(self.request)
        return access

    def get_queryset(self, **kwargs):
        self.entity = self.kwargs.get("entity")
        qs = (
            ContentType.objects.get(
                app_label__startswith="apis_", model=self.entity.lower()
            )
            .model_class()
            .objects.all()
        )
        self.filter = get_list_filter_of_entity(self.entity.title())(
            self.request.GET, queryset=qs
        )
        self.filter.form.helper = self.formhelper_class()
        return self.filter.qs

    def get_table(self, **kwargs):
        session = getattr(self.request, "session", False)
        entity = self.kwargs.get("entity")
        selected_cols = self.request.GET.getlist("columns")
        if session:
            edit_v = self.request.session.get("edit_views", False)
        else:
            edit_v = False
        if "table_fields" in settings.APIS_ENTITIES[entity.title()]:
            default_cols = settings.APIS_ENTITIES[entity.title()]["table_fields"]
        else:
            default_cols = ["name"]
        default_cols = default_cols + selected_cols
        self.table_class = get_entities_table(
            self.entity.title(), edit_v, default_cols=default_cols
        )
        table = super(GenericListViewNew, self).get_table()
        RequestConfig(
            self.request, paginate={"page": 1, "per_page": self.paginate_by}
        ).configure(table)
        return table

    def get_context_data(self, **kwargs):
        model = self.get_model()
        context = super(GenericListViewNew, self).get_context_data()
        context[self.context_filter_name] = self.filter
        context["entity"] = self.entity
        context["app_name"] = "apis_entities"
        entity = self.entity.title()
        context["entity_create_stanbol"] = GenericEntitiesStanbolForm(self.entity)
        if "browsing" in settings.INSTALLED_APPS:
            from browsing.models import BrowsConf

            context["conf_items"] = list(
                BrowsConf.objects.filter(model_name=self.entity).values_list(
                    "field_path", "label"
                )
            )
        context["docstring"] = "{}".format(model.__doc__)
        if model._meta.verbose_name_plural:
            context["class_name"] = "{}".format(model._meta.verbose_name.title())
        else:
            if model.__name__.endswith("s"):
                context["class_name"] = "{}".format(model.__name__)
            else:
                context["class_name"] = "{}s".format(model.__name__)
        try:
            context["get_arche_dump"] = model.get_arche_dump()
        except AttributeError:
            context["get_arche_dump"] = None
        try:
            context["create_view_link"] = model.get_createview_url()
        except AttributeError:
            context["create_view_link"] = None
        if "charts" in settings.INSTALLED_APPS:
            app_label = model._meta.app_label
            filtered_objs = ChartConfig.objects.filter(
                model_name=model.__name__.lower(), app_name=app_label
            )
            context["vis_list"] = filtered_objs
            context["property_name"] = self.request.GET.get("property")
            context["charttype"] = self.request.GET.get("charttype")
            if context["charttype"] and context["property_name"]:
                qs = self.get_queryset()
                chartdata = create_payload(
                    context["entity"],
                    context["property_name"],
                    context["charttype"],
                    qs,
                    app_label=app_label,
                )
                context = dict(context, **chartdata)
        try:
            context["enable_merge"] = settings.APIS_ENTITIES[entity.title()]["merge"]
        except KeyError:
            context["enable_merge"] = False
        try:
            togg_cols = settings.APIS_ENTITIES[entity.title()]["additional_cols"]
        except KeyError:
            togg_cols = []
        if context["enable_merge"] and self.request.user.is_authenticated:
            togg_cols = togg_cols + ["merge"]
        context["togglable_colums"] = togg_cols + ENTITIES_DEFAULT_COLS

        return context

    def render_to_response(self, context, **kwargs):
        download = self.request.GET.get("sep", None)
        if download and "browsing" in settings.INSTALLED_APPS:
            import datetime
            import time
            import pandas as pd

            sep = self.request.GET.get("sep", ",")
            timestamp = datetime.datetime.fromtimestamp(time.time()).strftime(
                "%Y-%m-%d-%H-%M-%S"
            )
            filename = "export_{}".format(timestamp)
            response = HttpResponse(content_type="text/csv")
            if context["conf_items"]:
                conf_items = context["conf_items"]
                try:
                    df = pd.DataFrame(
                        list(
                            self.get_queryset().values_list(*[x[0] for x in conf_items])
                        ),
                        columns=[x[1] for x in conf_items],
                    )
                except AssertionError:
                    response[
                        "Content-Disposition"
                    ] = 'attachment; filename="{}.csv"'.format(filename)
                    return response
            else:
                response[
                    "Content-Disposition"
                ] = 'attachment; filename="{}.csv"'.format(filename)
                return response
            if sep == "comma":
                df.to_csv(response, sep=",", index=False)
            elif sep == "semicolon":
                df.to_csv(response, sep=";", index=False)
            elif sep == "tab":
                df.to_csv(response, sep="\t", index=False)
            else:
                df.to_csv(response, sep=",", index=False)
            response["Content-Disposition"] = 'attachment; filename="{}.csv"'.format(
                filename
            )
            return response
        else:
            response = super(GenericListViewNew, self).render_to_response(context)
            return response


############################################################################
############################################################################
#
#   OtherViews
#
############################################################################
############################################################################


@user_passes_test(access_for_all_function)
def getGeoJson(request):
    """Used to retrieve GeoJsons for single objects"""
    # if request.is_ajax():
    pk_obj = request.GET.get("object_id")
    instance = get_object_or_404(Place, pk=pk_obj)
    uria = Uri.objects.filter(entity=instance)
    urib = UriCandidate.objects.filter(entity=instance)
    if urib.count() > 0:
        uric = urib
    elif uria.count() > 0:
        uric = uria
        add_info = ""
    lst_json = []
    if uric.count() > 0 and not instance.status.startswith("distinct"):
        for x in uric:
            o = retrieve_obj(x.uri)
            if o:
                url_r = reverse_lazy(
                    "apis:apis_entities:resolve_ambigue_place",
                    kwargs={
                        "pk": str(instance.pk),
                        "uri": o["representation"]["id"][7:],
                    },
                )
                select_text = "<a href='{}'>Select this URI</a>".format(url_r)
                try:
                    add_info = "<b>Confidence:</b> {}<br/><b>Feature:</b> <a href='{}'>{}</a>".format(
                        x.confidence, x.uri, x.uri
                    )
                except:
                    add_info = "<b>Confidence:</b>no value provided <br/><b>Feature:</b> <a href='{}'>{}</a>".format(
                        x.uri, x.uri
                    )
                r = {
                    "geometry": {
                        "type": "Point",
                        "coordinates": [
                            float(
                                o["representation"][
                                    "http://www.w3.org/2003/01/geo/wgs84_pos#long"
                                ][0]["value"]
                            ),
                            float(
                                o["representation"][
                                    "http://www.w3.org/2003/01/geo/wgs84_pos#lat"
                                ][0]["value"]
                            ),
                        ],
                    },
                    "type": "Feature",
                    "properties": {
                        "popupContent": "<b>ÖBL name:</b> %s<br/><b>Geonames:</b> %s<br/>%s<br/>%s"
                        % (
                            instance.name,
                            o["representation"][
                                "http://www.geonames.org/ontology#name"
                            ][0]["value"],
                            select_text,
                            add_info,
                        )
                    },
                    "id": x.pk,
                }
                lst_json.append(r)
    elif instance.lat is not None and instance.lng is not None:
        r = {
            "geometry": {"type": "Point", "coordinates": [instance.lng, instance.lat]},
            "type": "Feature",
            "properties": {"popupContent": "<b>Name:</b> %s<br/>" % (instance.name)},
            "id": instance.pk,
        }
        lst_json.append(r)

    return HttpResponse(json.dumps(lst_json), content_type="application/json")


@user_passes_test(access_for_all_function)
def getGeoJsonList(request):
    """Used to retrieve a list of GeoJsons. To generate the list the kind of connection
    and the connected entity is needed"""
    relation = AbstractRelation.get_relation_class_of_name(request.GET.get("relation"))
    # relation_type = request.GET.get("relation_type")
    objects = relation.objects.filter(related_place__status="distinct").select_related(
        "related_person", "related_place", "relation_type"
    )
    lst_json = []
    for x in objects:
        pers_url = x.related_person.get_absolute_url()
        place_url = x.related_place.get_absolute_url()
        r = {
            "geometry": {
                "type": "Point",
                "coordinates": [x.related_place.lng, x.related_place.lat],
            },
            "type": "Feature",
            "relation_type": x.relation_type.name,
            "properties": {
                "popupContent": "<b>Person:</b> <a href='%s'>%s</a><br/><b>Connection:</b> %s<br/><b>Place:</b> <a href='%s'>%s</a>"
                % (
                    pers_url,
                    x.related_person,
                    x.relation_type,
                    place_url,
                    x.related_place,
                )
            },
            "id": x.pk,
        }
        lst_json.append(r)
    return HttpResponse(json.dumps(lst_json), content_type="application/json")


@user_passes_test(access_for_all_function)
def getNetJsonList(request):
    """Used to retrieve a Json to draw a network"""
    relation = AbstractRelation.get_relation_class_of_name("PersonPlace")
    objects = relation.objects.filter(related_place__status="distinct")
    nodes = dict()
    edges = []

    for x in objects:
        if x.related_place.pk not in nodes.keys():
            place_url = reverse_lazy(
                "apis:apis_entities:place_edit", kwargs={"pk": str(x.related_place.pk)}
            )
            tt = (
                "<div class='arrow'></div>\
            <div class='sigma-tooltip-header'>%s</div>\
            <div class='sigma-tooltip-body'>\
            <table>\
                <tr><th>Type</th> <td>%s</td></tr>\
                <tr><th>Entity</th> <td><a href='%s'>Link</a></td></tr>\
            </table>\
            </div>"
                % (x.related_place.name, "place", place_url)
            )
            nodes[x.related_place.pk] = {
                "type": "place",
                "label": x.related_place.name,
                "id": str(x.related_place.pk),
                "tooltip": tt,
            }
        if x.related_person.pk not in nodes.keys():
            pers_url = reverse_lazy(
                "apis:apis_entities:person_edit",
                kwargs={"pk": str(x.related_person.pk)},
            )
            tt = (
                "<div class='arrow'></div>\
            <div class='sigma-tooltip-header'>%s</div>\
            <div class='sigma-tooltip-body'>\
            <table>\
                <tr><th>Type</th> <td>%s</td></tr>\
                <tr><th>Entity</th> <td><a href='%s'>Link</a></td></tr>\
            </table>\
            </div>"
                % (str(x.related_person), "person", pers_url)
            )
            nodes[x.related_person.pk] = {
                "type": "person",
                "label": str(x.related_person),
                "id": str(x.related_person.pk),
                "tooltip": tt,
            }
        edges.append(
            {
                "source": x.related_person.pk,
                "target": x.related_place.pk,
                "kind": x.relation_type.name,
                "id": str(x.pk),
            }
        )
    lst_json = {"edges": edges, "nodes": [nodes[x] for x in nodes.keys()]}

    return HttpResponse(json.dumps(lst_json), content_type="application/json")


@user_passes_test(access_for_all_function)
def getNetJsonListInstitution(request):
    """Used to retrieve a Json to draw a network"""
    relation = AbstractRelation.get_relation_class_of_name("PersonInstitution")
    objects = relation.objects.all()
    nodes = dict()
    edges = []

    for x in objects:
        if x.related_institution.pk not in nodes.keys():
            inst_url = reverse_lazy(
                "apis:apis_entities:institution_edit",
                kwargs={"pk": str(x.related_institution.pk)},
            )
            tt = (
                "<div class='arrow'></div>\
            <div class='sigma-tooltip-header'>%s</div>\
            <div class='sigma-tooltip-body'>\
            <table>\
                <tr><th>Type</th> <td>%s</td></tr>\
                <tr><th>Entity</th> <td><a href='%s'>Link</a></td></tr>\
            </table>\
            </div>"
                % (x.related_institution.name, "institution", inst_url)
            )
            nodes[x.related_institution.pk] = {
                "type": "institution",
                "label": x.related_institution.name,
                "id": str(x.related_institution.pk),
                "tooltip": tt,
            }
        if x.related_person.pk not in nodes.keys():
            pers_url = reverse_lazy(
                "apis:apis_entities:person_edit",
                kwargs={"pk": str(x.related_person.pk)},
            )
            tt = (
                "<div class='arrow'></div>\
            <div class='sigma-tooltip-header'>%s</div>\
            <div class='sigma-tooltip-body'>\
            <table>\
                <tr><th>Type</th> <td>%s</td></tr>\
                <tr><th>Entity</th> <td><a href='%s'>Link</a></td></tr>\
            </table>\
            </div>"
                % (str(x.related_person), "person", pers_url)
            )
            nodes[x.related_person.pk] = {
                "type": "person",
                "label": str(x.related_person),
                "id": str(x.related_person.pk),
                "tooltip": tt,
            }
        edges.append(
            {
                "source": x.related_person.pk,
                "target": x.related_institution.pk,
                "kind": x.relation_type.name,
                "id": str(x.pk),
            }
        )
    lst_json = {"edges": edges, "nodes": [nodes[x] for x in nodes.keys()]}

    return HttpResponse(json.dumps(lst_json), content_type="application/json")


@login_required
def resolve_ambigue_place(request, pk, uri):
    """Only used to resolve place names."""
    with reversion.create_revision():
        uri = "http://" + uri
        entity = Place.objects.get(pk=pk)
        pl_n = RDFParser(uri, kind="Place")
        pl_n.create_objct()
        pl_n_1 = pl_n.save()
        pl_n_1 = pl_n.merge(entity)
        url = pl_n_1.get_absolute_url()
        if pl_n.created:
            pl_n_1.status = "distinct (manually resolved)"
            pl_n_1.save()
        UriCandidate.objects.filter(entity=entity).delete()
        reversion.set_user(request.user)

    return HttpResponseRedirect(url)


@login_required
def resolve_ambigue_person(request):
    if request.method == "POST":
        form = PersonResolveUriForm(request.POST)
    if form.is_valid():
        pers = form.save()
        return redirect(
            reverse("apis:apis_entities:person_edit", kwargs={"pk": pers.pk})
        )


############################################################################
############################################################################
#
#   VisualizationViews
#
############################################################################
############################################################################


@user_passes_test(access_for_all_function)
def birth_death_map(request):
    return render(request, "apis:apis_entities/map_list.html")


@user_passes_test(access_for_all_function)
def pers_place_netw(request):
    return render(request, "apis:apis_entities/network.html")


@user_passes_test(access_for_all_function)
def pers_inst_netw(request):
    return render(request, "apis:apis_entities/network_institution.html")


@user_passes_test(access_for_all_function)
def generic_network_viz(request):
    if request.method == "GET":
        form = NetworkVizFilterForm()
        return render(
            request,
            "apis:apis_entities/generic_network_visualization.html",
            {"form": form},
        )


############################################################################
############################################################################
#
#  Reversion Views
#
############################################################################
############################################################################
# TODO: add again as soon as the module has been bumped to new django version
# class ReversionCompareView(HistoryCompareDetailView):
#    template_name = 'apis_entities/compare_base.html'

#    def dispatch(self, request, app, kind, pk, *args, **kwargs):
#        self.model = ContentType.objects.get(app_label=app, model=kind).model_class()
#        return super(ReversionCompareView, self).dispatch(request, *args, **kwargs)



from functools import reduce
import copy
import importlib
import inspect

from django.conf import settings

# from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from rest_framework import pagination, serializers, viewsets
from rest_framework import renderers
from rest_framework.response import Response
from drf_spectacular.contrib.django_filters import (
    DjangoFilterBackend as DjangoFilterbackendSpectacular,
)
from drf_spectacular.utils import (
    extend_schema,
    extend_schema_field,
    OpenApiParameter,
    extend_schema_serializer,
)
from drf_spectacular.types import OpenApiTypes
from django import forms
from django_filters import rest_framework as filters
from .apis_metainfo.models import TempEntityClass
from .api_renderers import NetJsonRenderer
from .apis_relations.models import AbstractRelation
from apis_core.helper_functions.ContentType import GetContentTypes


if "apis_highlighter" in getattr(settings, "INSTALLED_APPS"):
    from apis_highlighter.highlighter import highlight_text_new
    from apis_highlighter.serializer import annotationSerializer
    from apis_highlighter.models import Annotation


try:
    MAX_AGE = settings.MAX_AGE
except AttributeError:
    MAX_AGE = 0


def deep_get(dictionary, keys, default=None):
    return reduce(
        lambda d, key: d.get(key, default) if isinstance(d, dict) else default,
        keys.split("."),
        dictionary,
    )


def create_query_parameters(entity):
    print(entity)
    for f in entity._meta.fields:
        print(f.name, f.__class__.__name__)


class CustomPagination(pagination.LimitOffsetPagination):
    def get_paginated_response(self, data):
        return Response(
            {
                "next": self.get_next_link(),
                "previous": self.get_previous_link(),
                "count": self.count,
                "limit": self.limit,
                "offset": self.offset,
                "results": data,
            }
        )


class ApisBaseSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    label = serializers.SerializerMethodField(method_name="add_label")
    url = serializers.SerializerMethodField(method_name="add_uri")

    @extend_schema_field(OpenApiTypes.INT)
    def add_id(self, obj):
        return obj.pk

    @extend_schema_field(OpenApiTypes.STR)
    def add_label(self, obj):
        return str(obj)

    @extend_schema_field(OpenApiTypes.URI)
    def add_uri(self, obj):
        return self.context["view"].request.build_absolute_uri(
            reverse(
                "apis:apis_api:{}-detail".format(obj.__class__.__name__.lower()),
                kwargs={"pk": obj.pk},
            )
        )

    @extend_schema_field(OpenApiTypes.OBJECT)
    def add_type(self, obj):
        lst_type = ["kind", "type", "collection_type", "relation_type"]
        lst_kind = [
            x
            for x in obj._meta.fields
            if x.name in lst_type and "apis_vocabularies" in str(x.related_model)
        ]
        if len(lst_kind):
            pk_obj = getattr(obj, f"{lst_kind[0].name}_id")
            if pk_obj is not None:
                obj_type = getattr(obj, str(lst_kind[0].name))
                res = {
                    "id": pk_obj,
                    "url": self.context["view"].request.build_absolute_uri(
                        reverse(
                            f"apis:apis_api:{lst_kind[0].related_model.__name__.lower()}-detail",
                            kwargs={"pk": pk_obj},
                        )
                    ),
                    "type": obj_type.__class__.__name__,
                    "label": str(obj_type),
                    "parent_class": getattr(obj, "parent_class_id", None),
                }
                return res

    class Meta:
        model = TempEntityClass
        fields = ["id", "label", "url"]


class EntitySerializer(ApisBaseSerializer):
    type = serializers.SerializerMethodField(method_name="add_type")

    def add_type(self, instance):
        return instance.__class__.__name__

    class Meta(ApisBaseSerializer.Meta):
        fields = ApisBaseSerializer.Meta.fields + ["type"]


class LabelSerializer(ApisBaseSerializer):
    parent_id = serializers.PrimaryKeyRelatedField(
        many=False, source="parent_class_id", read_only=True
    )

    @extend_schema_field(OpenApiTypes.INT)
    def add_parent_id(self, obj):
        return obj.parent_class_id

    class Meta(ApisBaseSerializer.Meta):
        fields = ApisBaseSerializer.Meta.fields + [
            "parent_id",
        ]


class VocabsBaseSerializer(LabelSerializer, EntitySerializer):
    pass


class RelationObjectSerializer2(ApisBaseSerializer):
    relation_type = VocabsBaseSerializer(read_only=True)
    related_entity = serializers.SerializerMethodField(method_name="add_related_entity")

    @extend_schema_field(EntitySerializer)
    def add_related_entity(self, instance):
        for at in dir(instance):
            if (
                at.startswith("related_")
                and at.endswith("_id")
                and getattr(instance, at) != self._pk_instance
            ):
                return EntitySerializer(
                    getattr(instance, at[:-3]), context=self.context
                ).data

    class Meta(ApisBaseSerializer.Meta):
        fields = ApisBaseSerializer.Meta.fields + ["relation_type", "related_entity"]

    def __init__(self, *args, **kwargs):
        self._pk_instance = kwargs.pop("pk_instance")
        super(RelationObjectSerializer2, self).__init__(*args, **kwargs)


if "apis_highlighter" in getattr(settings, "INSTALLED_APPS"):

    class AnnotationSerializer(serializers.ModelSerializer):
        related_object = VocabsBaseSerializer(
            source="get_related_entity", read_only=True, many=False
        )

        class Meta:
            model = Annotation
            fields = ["id", "start", "end", "related_object"]

def generic_serializer_creation_factory():
    lst_cont = GetContentTypes().get_model_classes()
    not_allowed_filter_fields = [
        "useradded",
        "vocab_name",
        "parent_class",
        "vocab",
        "entity",
        "autofield",
    ]
    for cont in lst_cont:
        prefetch_rel = []
        select_related = []
        test_search = getattr(settings, cont.__module__.split(".")[1].upper(), False)
        entity_str = str(cont.__name__).replace(" ", "")
        entity = cont
        app_label = cont.__module__.split(".")[1].lower()
        exclude_lst = []
        if app_label == "apis_entities":
            exclude_lst = deep_get(test_search, "{}.api_exclude".format(entity_str), [])
        else:
            set_prem = getattr(settings, cont.__module__.split(".")[1].upper(), {})
            exclude_lst = deep_get(set_prem, "exclude", [])
            exclude_lst.extend(deep_get(set_prem, "{}.exclude".format(entity_str), []))
        entity_field_name_list = []
        for x in entity._meta.get_fields():
            entity_field_name_list.append(x.name)
        exclude_lst_fin = []
        for x in exclude_lst:
            if x in entity_field_name_list:
                exclude_lst_fin.append(x)
        if entity_str.lower() == "text":
            exclude_lst_fin.extend(["kind", "source"])
        if app_label == "apis_relations":
            exclude_lst_fin.extend(["text", "collection"])
        for f in entity._meta.get_fields():
            if f.__class__.__name__ == "ManyToManyField":
                prefetch_rel.append(f.name)
            elif f.__class__.__name__ == "ForeignKey":
                select_related.append(f.name)

        class TemplateSerializer(serializers.HyperlinkedModelSerializer):

            id = serializers.ReadOnlyField()
            url = serializers.HyperlinkedIdentityField(view_name=f"apis:apis_api:{entity_str.lower()}-detail")
            _entity = entity
            _exclude_lst = exclude_lst_fin
            _app_label = app_label

            class Meta:

                model = entity
                exclude = exclude_lst_fin

            def add_labels(self, obj):
                return {"id": obj.pk, "label": str(obj)}

            if entity_str.lower() == "text":

                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self._highlight = False
                    self.fields["kind"] = LabelSerializer(many=False, read_only=True)

            else:

                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    entity_str = self._entity.__name__
                    app_label = self._app_label
                    lst_labels_set = deep_get(
                        getattr(settings, app_label.upper(), {}),
                        "{}.labels".format(entity_str),
                        [],
                    )
                    for f in self._entity._meta.get_fields():
                        if getattr(settings, "APIS_API_EXCLUDE_SETS", False) and str(
                            f.name
                        ).endswith("_set"):
                            if f.name in self.fields.keys():
                                self.fields.pop(f.name)
                            continue
                        ck_many = f.__class__.__name__ == "ManyToManyField"
                        if f.name in self._exclude_lst:
                            continue
                        elif (
                            f.__class__.__name__
                            in [
                                "ManyToManyField",
                                "ForeignKey",
                            ]
                            and "apis_vocabularies" not in str(f.related_model)
                        ):
                            self.fields[f.name] = ApisBaseSerializer(
                                many=ck_many, read_only=True
                            )
                        elif f.__class__.__name__ in ["ManyToManyField", "ForeignKey"]:
                            self.fields[f.name] = LabelSerializer(many=ck_many, read_only=True)

        TemplateSerializer.__name__ = TemplateSerializer.__qualname__ = f"{entity_str.title().replace(' ', '')}Serializer"

        class TemplateSerializerRetrieve(TemplateSerializer):

            if entity_str.lower() == "text":
                text = serializers.SerializerMethodField(method_name="txt_serializer_add_text")
                if "apis_highlighter" in getattr(settings, "INSTALLED_APPS"):
                    annotations = serializers.SerializerMethodField(method_name="txt_serializer_add_annotations")

                    @extend_schema_field(AnnotationSerializer(many=True))
                    def txt_serializer_add_annotations(self, instance):
                        if self._highlight:
                            return AnnotationSerializer(
                                self._annotations, context=self.context, many=True
                            ).data
                        else:
                            return None

                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    highlight = self.context.get("highlight", True)
                    self._inline_annotations = False
                    if highlight is not None and "apis_highlighter" in getattr(
                        settings, "INSTALLED_APPS"
                    ):
                        self._highlight = highlight
                        if self._highlight == "":
                            self._highlight = True
                        if not isinstance(self._highlight, bool):
                            if self._highlight.lower() == "false":
                                self._highlight = False
                        self._ann_proj_pk = self.context.get("ann_proj_pk", None)
                        self._types = self.context.get("types", None)
                        self._users_show = self.context.get("users_show", None)
                        self._inline_annotations = self.context.get("inline_annotations", True)
                        if not isinstance(self._inline_annotations, bool):
                            if self._inline_annotations.lower() == "false":
                                self._inline_annotations = False
                            elif self._inline_annotations.lower() == "true":
                                self._inline_annotations = True
                        try:
                            self._txt_html, self._annotations = highlight_text_new(
                                self.instance,
                                set_ann_proj=self._ann_proj_pk,
                                types=self._types,
                                users_show=self._users_show,
                                inline_annotations=self._inline_annotations,
                            )
                            qs_an = {"text": self.instance}
                            if self._users_show is not None:
                                qs_an["users_added__in"] = self._users_show
                            if self._ann_proj_pk is not None:
                                qs_an["annotation_project_id"] = self._ann_proj_pk
                            #self._annotations = Annotation.objects.filter(
                            #    **qs_an
                            #)  # FIXME: Currently this QS is called twice (highlight_text_new)
                        except Exception as e:
                            self._txt_html = ""
                            self._annotations = []
                    else:
                        self._highlight = False
                    self.fields["kind"] = LabelSerializer(many=False, read_only=True)

                def txt_serializer_add_text(self, instance):
                    if self._inline_annotations:
                        return self._txt_html
                    else:
                        return instance.text

            else:

                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    entity_str = self._entity.__name__
                    app_label = self._app_label
                    lst_labels_set = deep_get(
                        getattr(settings, app_label.upper(), {}),
                        "{}.labels".format(entity_str),
                        [],
                    )
                    for f in self._entity._meta.get_fields():
                        if getattr(settings, "APIS_API_EXCLUDE_SETS", False) and str(
                            f.name
                        ).endswith("_set"):
                            if f.name in self.fields.keys():
                                self.fields.pop(f.name)
                            continue
                        ck_many = f.__class__.__name__ == "ManyToManyField"
                        if f.name in self._exclude_lst:
                            continue
                        elif (
                            f.__class__.__name__
                            in [
                                "ManyToManyField",
                                "ForeignKey",
                            ]
                            and "apis_vocabularies" not in str(f.related_model)
                        ):
                            self.fields[f.name] = ApisBaseSerializer(
                                many=ck_many, read_only=True
                            )
                        elif f.__class__.__name__ in ["ManyToManyField", "ForeignKey"]:
                            self.fields[f.name] = LabelSerializer(many=ck_many, read_only=True)
                    # include = list(ContentType.objects.filter(app_label="apis_relations", model__icontains=entity_str).values_list('model', flat=True))
                    include = [
                        x
                        for x in lst_cont
                        if x.__module__ == "apis_core.apis_relations.models"
                        and entity_str.lower() in x.__name__.lower()
                    ]
                    if len(include) > 0 and len(args) > 0:
                        inst_pk2 = args[0].pk
                        self.fields["relations"] = RelationObjectSerializer2(
                            read_only=True,
                            source="get_related_relation_instances",
                            many=True,
                            pk_instance=inst_pk2,
                        )

        TemplateSerializerRetrieve.__name__ = TemplateSerializerRetrieve.__qualname__ = f"{entity_str.title().replace(' ', '')}DetailSerializer"

        allowed_fields_filter = {
            "IntegerField": ["in", "range", "exact"],
            "CharField": ["exact", "icontains", "iregex", "isnull"],
            "DateField": ["year", "lt", "gt", "year__lt", "year__gt", "exact"],
            "PositiveIntegerField": ["in", "range", "exact"],
            "AutoField": ["in", "exact"],
        }
        filterset_dict = {}
        filter_fields = {}

        for field in entity._meta.fields + entity._meta.many_to_many:
            if getattr(settings, "APIS_API_EXCLUDE_SETS", False) and "_set" in str(
                field.name.lower()
            ):
                continue
            if (
                field.name.lower() in not_allowed_filter_fields
                or field.name == "tempentityclass_ptr"
            ):
                continue
            elif field.__class__.__name__ in ["ForeignKey", "ManyToManyField"]:
                filter_fields[field.name] = ["exact"]
                if field.__class__.__name__ == "ForeignKey":
                    filter_fields[field.name].append("in")
                for f2 in field.related_model._meta.fields:
                    if f2.__class__.__name__ in [
                        "CharField",
                        "DateField",
                        "IntegerField",
                        "AutoField",
                    ]:
                        filter_fields[
                            f"{field.name}__{f2.name}"
                        ] = allowed_fields_filter[f2.__class__.__name__]
                continue
            if field.__class__.__name__ in allowed_fields_filter.keys():
                filter_fields[field.name] = allowed_fields_filter[
                    field.__class__.__name__
                ]
            else:
                filter_fields[field.name] = ["exact"]
        additional_filters = getattr(settings, "APIS_API_ADDITIONAL_FILTERS", False)
        if additional_filters:
            if entity_str in additional_filters.keys():
                for f1 in additional_filters[entity_str]:
                    if f1[0] not in filter_fields.keys():
                        filter_fields[f1[0]] = f1[1]

        class MetaFilter(object):

            model = entity
            fields = filter_fields

        filterset_dict["Meta"] = MetaFilter

        class TemplateViewSet(viewsets.ModelViewSet):

            _select_related = select_related
            _prefetch_rel = prefetch_rel
            pagination_class = CustomPagination
            model = entity
            filter_backends = (DjangoFilterbackendSpectacular,)
            filterset_fields = filter_fields
            depth = 2
            renderer_classes = (
                renderers.JSONRenderer,
                renderers.BrowsableAPIRenderer,
                NetJsonRenderer,
            )
            _serializer_class = TemplateSerializer
            _serializer_class_retrieve = TemplateSerializerRetrieve

            def get_serializer_class(self, *arg, **kwargs):
                if self.action == "list":
                    return self._serializer_class
                else:
                    return self._serializer_class_retrieve

            def get_serializer_context(self):
                context = super(self.__class__, self).get_serializer_context()
                if self.action == "retrieve" and self.model.__name__.lower() == "text":
                    cont = {}
                    cont["highlight"] = self.request.query_params.get("highlight", None)
                    cont["ann_proj_pk"] = self.request.query_params.get("ann_proj_pk", None)
                    cont["types"] = self.request.query_params.get("types", None)
                    cont["users_show"] = self.request.query_params.get("users_show", None)
                    cont["inline_annotations"] = self.request.query_params.get(
                        "inline_annotations", True
                    )
                    context.update(cont)
                return context

            def get_queryset(self):
                if "apis_relations" in str(self.model):
                    qs = self.model.objects.filter_for_user()
                else:
                    qs = self.model.objects.all()
                if len(self._prefetch_rel) > 0:
                    qs = qs.prefetch_related(*self._prefetch_rel)
                if len(self._select_related) > 0:
                    qs = qs.select_related(*self._select_related)
                return qs

            @extend_schema(responses=TemplateSerializer(many=True))
            def list_viewset(self, request):
                res = super(self.__class__, self).list(request)
                return res

            def dispatch(self, request, *args, **kwargs):
                return super(self.__class__, self).dispatch(request, *args, **kwargs)

            if entity_str.lower() == "text":
                @extend_schema(
                    parameters=[
                        OpenApiParameter(
                            name="highlight",
                            description="Whether to add annotations or not, defaults to true",
                            type=OpenApiTypes.BOOL,
                        ),
                        OpenApiParameter(
                            name="inline_annotations",
                            description="Whether to add html5 mark tags for annotations to the text, defaults to false",
                            type=OpenApiTypes.BOOL,
                        ),
                        OpenApiParameter(
                            name="ann_proj_pk",
                            description="PK of the annotation project to use for annotations",
                            type=OpenApiTypes.INT,
                        ),
                        OpenApiParameter(
                            name="types",
                            description="Content type pks of annotation types to show. E.g. PersonPlace relations (comma sperated list)",
                            type=OpenApiTypes.STR,
                        ),
                        OpenApiParameter(
                            name="users_show",
                            description="Filter annotations for users. PKs of users, comma seperated list",
                            type=OpenApiTypes.STR,
                        ),
                    ],
                    responses={200: TemplateSerializerRetrieve},
                )
                def retrieve(self, request, pk=None):
                    res = super(self.__class__, self).retrieve(request, pk=pk)
                    return res

        TemplateViewSet.__name__ = TemplateViewSet.__qualname__ = f"Generic{entity_str.title().replace(' ', '')}ViewSet"

        serializers_dict[TemplateSerializer.__name__] = TemplateSerializer
        views[f"{entity_str.lower().replace(' ', '')}"] = TemplateViewSet

serializers_dict = dict()
views = dict()
# filter_classes = dict()
# lst_filter_classes_check = []
generic_serializer_creation_factory()



from rest_framework.authentication import TokenAuthentication

class TokenAuthSupportQueryString(TokenAuthentication):
    """
    Extend the TokenAuthentication class to support querystring authentication
    in the form of "http://www.example.com/?auth_token=<token_key>"
    """
    def authenticate(self, request):
        # Check if 'token_auth' is in the request query params.
        # Give precedence to 'Authorization' header.
        if 'api_key' in request.query_params and \
                        'HTTP_AUTHORIZATION' not in request.META:
            return self.authenticate_credentials(request.query_params.get('api_key'))
        else:
            return super(TokenAuthSupportQueryString, self).authenticate(request)


from dal import autocomplete
from django.apps import apps
from django.conf import settings
from django.contrib import admin
from django.urls import reverse
from csvexport.actions import csvexport


class BaseAdminVocabularies(admin.ModelAdmin):
    '''Base class used to store the User foreign key in the background when someone adds a vocab.'''

    search_fields = ('name', 'parent_class__name')
    exclude = ('userAdded',)
    actions = [csvexport]

    def get_fields(self, request, obj=None):
        lst = super(BaseAdminVocabularies, self).get_fields(request, obj=None)
        if not request.user.is_superuser:
            lst.remove('status')
        return lst

    def get_queryset(self, request):
        qs = super(BaseAdminVocabularies, self).get_queryset(request)
        if request.user.is_superuser:
            return qs.all()
        return qs.filter(userAdded__groups__in=request.user.groups.all()).distinct()

    def save_model(self, request, obj, form, change):
        if not change:
            obj.userAdded = request.user
        obj.save()

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        attrs = {'data-placeholder': 'Type to get suggestions',
                 'data-minimum-input-length': getattr(settings, "APIS_MIN_CHAR", 3),
                 'data-html': True}
        c_name = db_field.model.__name__
        qs = super(BaseAdminVocabularies, self).get_queryset(request)
        if c_name.endswith('Relation') and db_field.name == 'parent_class':
            qs = db_field.model
        if db_field.name == "parent_class" and request.user.is_superuser:
            kwargs["queryset"] = qs.all()
        elif db_field.name == "parent_class":
            kwargs["queryset"] = qs.filter(userAdded__groups__in=request.user.groups.all())
        kwargs['widget'] = autocomplete.Select2(
            url=reverse(
                'apis:apis_vocabularies:generic_vocabularies_autocomplete',
                kwargs={
                    'vocab': self.model.__name__.lower(),
                    'direct': 'normal'
                }
            ), attrs=attrs)

        return super(BaseAdminVocabularies, self).formfield_for_foreignkey(
            db_field, request, **kwargs
        )


class VocabsRelationAdmin(BaseAdminVocabularies):
    list_display = ('name', 'label')
    search_fields = ('name', 'parent_class__name')



app = apps.get_app_config('apis_vocabularies')


for model_name, model in app.models.items():

    if model_name.endswith('relation'):
        admin.site.register(model, VocabsRelationAdmin)
    else:
        admin.site.register(model, BaseAdminVocabularies)



# Generated by Django 2.2.12 on 2020-05-29 14:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apis_metainfo', '0002_auto_20200121_1227'),
    ]

    operations = [
        migrations.AddField(
            model_name='tempentityclass',
            name='published',
            field=models.BooleanField(default=False),
        ),
    ]



import fnmatch
import os
import re
from datetime import datetime
from io import TextIOWrapper

import requests
from django.conf import settings
from django.db.models import Q, Prefetch
from django.http import Http404
from django.shortcuts import redirect
from django.urls import reverse
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, viewsets
from rest_framework.decorators import api_view
from rest_framework.generics import GenericAPIView
from rest_framework.pagination import PageNumberPagination
from rest_framework.parsers import FileUploadParser
from rest_framework.permissions import DjangoObjectPermissions
from rest_framework.response import Response
from rest_framework.reverse import reverse_lazy
from rest_framework.settings import api_settings
from rest_framework.views import APIView

from apis_core.apis_metainfo.api_renderers import PaginatedCSVRenderer
from apis_core.apis_metainfo.models import TempEntityClass, Uri
from apis_core.apis_relations.models import (
    PersonPlace,
    InstitutionPlace,
    AbstractRelation,
    PersonInstitution,
)
from apis_core.apis_vocabularies.models import (
    VocabsBaseClass,
    PersonPlaceRelation,
    InstitutionPlaceRelation,
)
from apis_core.default_settings.NER_settings import autocomp_settings, stb_base
from apis_core.helper_functions.RDFParser import RDFParser
from apis_core.helper_functions.stanbolQueries import find_loc
from .api_renderers import (
    EntityToTEI,
    EntityToCIDOCXML,
    EntityToProsopogrAPhI,
    EntityToCIDOCN3,
    EntityToCIDOCNQUADS,
    EntityToCIDOCTURTLE,
)
from .models import Event, Institution, Person, Place, Work, AbstractEntity
from .serializers import (
    EventSerializer,
    GeoJsonSerializer,
    InstitutionSerializer,
    NetJsonEdgeSerializer,
    NetJsonNodeSerializer,
    PersonSerializer,
    PlaceSerializer,
    WorkSerializer,
    GeoJsonSerializerTheme,
    LifePathSerializer,
)
from .serializers_generic import EntitySerializer


# from metainfo.models import TempEntityClass


class StandardResultsSetPagination(PageNumberPagination):
    page_size = 25
    page_size_query_param = "page_size"
    max_page_size = 1000


class GetEntityGeneric(GenericAPIView):
    serializer_class = EntitySerializer
    queryset = TempEntityClass.objects.all()
    renderer_classes = tuple(api_settings.DEFAULT_RENDERER_CLASSES) + (
        EntityToTEI,
        EntityToCIDOCXML,
        EntityToProsopogrAPhI,
        EntityToCIDOCN3,
        EntityToCIDOCNQUADS,
        EntityToCIDOCTURTLE,
    )
    if getattr(settings, "APIS_RENDERERS", None) is not None:
        rend_add = tuple()
        for rd in settings.APIS_RENDERERS:
            rend_mod = __import__(rd)
            for name, cls in rend_mod.__dict__.items():
                rend_add + (cls,)
        renderer_classes += rend_add

    def get_object(self, pk, request):
        try:
            return TempEntityClass.objects_inheritance.get_subclass(pk=pk)
        except TempEntityClass.DoesNotExist:
            uri2 = Uri.objects.filter(uri=request.build_absolute_uri())
            if uri2.count() == 1:
                return TempEntityClass.objects_inheritance.get_subclass(
                    pk=uri2[0].entity_id
                )
            else:
                raise Http404

    def get(self, request, pk):
        ent = self.get_object(pk, request)
        res = EntitySerializer(ent, context={"request": request})
        return Response(res.data)


@api_view(["GET"])
def uri_resolver(request):
    uri = request.query_params.get("uri", None)
    f = request.query_params.get("target_format", "gui")
    if uri is None:
        raise Http404
    else:
        uri = Uri.objects.get(uri=uri)
        if f == "gui":
            ent = TempEntityClass.objects_inheritance.get_subclass(pk=uri.entity_id)
            c_name = ent.__class__.__name__
            url = reverse(
                "apis_core:apis_entities:generic_entities_detail_view",
                kwargs={"pk": uri.entity_id, "entity": c_name.lower()},
            )
        else:
            url = reverse(
                "apis_core:apis_api2:GetEntityGeneric", kwargs={"pk": uri.entity_id}
            ) + "?format={}".format(f)
        return redirect(url)


class InstitutionViewSet(viewsets.ModelViewSet):
    """Serialization of the institution class.
    In addition to the institution this view includes related texts and \
    the kind of the institution (separated object)."""

    permission_classes = (DjangoObjectPermissions,)
    queryset = Institution.objects.all()
    serializer_class = InstitutionSerializer
    filter_backends = (DjangoFilterBackend, filters.SearchFilter)
    depth = 2
    filter_fields = ("name", "kind__name", "collection__name")
    search_fields = ("name",)


class PersonViewSet(viewsets.ModelViewSet):
    """Serialization of the Person class.
    In addition to the person this view includes related texts (e.g. biographies), \
    a list of professions (separated objects) and the collection it belongs to."""

    permission_classes = (DjangoObjectPermissions,)
    queryset = Person.objects.all()
    serializer_class = PersonSerializer
    depth = 2
    filter_backends = (DjangoFilterBackend, filters.SearchFilter)
    filter_fields = (
        "name",
        "first_name",
        "gender",
        "profession__name",
        "collection__name",
        "uri__uri",
    )
    search_fields = ("name", "first_name")


class PlaceViewSet(viewsets.ModelViewSet):
    """Serialization of the Place class.
    In addition to the place this view includes related texts (e.g. from the ÖBL database), \
     and the collection it belongs to."""

    permission_classes = (DjangoObjectPermissions,)
    queryset = Place.objects.all()
    serializer_class = PlaceSerializer
    filter_backends = (DjangoFilterBackend, filters.SearchFilter)
    depth = 2
    filter_fields = ("name", "kind__name", "collection__name")
    search_fields = ("name",)
    renderer_classes = tuple(api_settings.DEFAULT_RENDERER_CLASSES) + (
        PaginatedCSVRenderer,
    )


class EventViewSet(viewsets.ModelViewSet):
    """Serialization of the event class.
    In addition to the event this view includes related texts and
    the kind of the event (separated object)."""

    permission_classes = (DjangoObjectPermissions,)
    queryset = Event.objects.all()
    serializer_class = EventSerializer
    filter_backends = (DjangoFilterBackend, filters.SearchFilter)
    depth = 2
    filter_fields = ("name", "kind__name", "collection__name")
    search_fields = ("name",)


class WorkViewSet(viewsets.ModelViewSet):
    """Serialization of the work class.
    In addition to the work this view includes related texts and
    the kind of the work (separated object)."""

    permission_classes = (DjangoObjectPermissions,)
    queryset = Work.objects.all()
    serializer_class = WorkSerializer
    filter_backends = (DjangoFilterBackend, filters.SearchFilter)
    depth = 2
    filter_fields = ("name", "kind__name", "collection__name")
    search_fields = ("name",)


class PlaceGeoJsonViewSet(viewsets.ViewSet):
    def list(self, request):
        queryset = request.query_params.get("q", None)
        diff = re.match(r"^([^\[]*)\[(\d+)\]", queryset)
        match_url = re.search(r"geonames[^0-9]+([0-9]+)", queryset)
        res = False
        if match_url:
            for u in autocomp_settings["Place"]:
                if "url" in u.keys():
                    site = re.search("site/([^/]+)/", u["url"])
                    if site:
                        params = {
                            "id": "http://sws.geonames.org/{}/".format(
                                match_url.group(1)
                            )
                        }
                        headers = {"Content-Type": "application/json"}
                        w = requests.get(
                            stb_base + site.group(1) + "/entity",
                            params=params,
                            headers=headers,
                        )
                        if w.status_code == 200:
                            if "representation" in w.json().keys():
                                res = [True, w.json()["representation"]]
                                break
        elif diff and queryset:
            queryset = diff.group(1)
            res = find_loc(queryset.split(","), dec_diff=int(diff.group(2)))
        elif queryset:
            res = find_loc(queryset.split(","))
        else:
            res = False, False
        if res[1]:
            if type(res[1]) is not list:
                res = (res[0], [res[1]])
            serializer = GeoJsonSerializer(
                res[1],
                many=True,
                context={"p_pk": request.query_params.get("p_pk", None)},
            )
        else:
            serializer = GeoJsonSerializer(
                [], many=True, context={"p_pk": request.query_params.get("p_pk", None)}
            )
        return Response(serializer.data)


class NetJsonViewSet(viewsets.GenericViewSet):
    def get_queryset(self):
        rel = self.request.data["select_relation"]
        q = AbstractRelation.get_relation_class_of_name("".join(rel.split("-")))
        return q.objects.all()

    def create(self, request):
        rel = request.data["select_relation"]
        if "select_kind" in request.data.keys():
            kind = request.data["select_kind"]
        else:
            kind = None
        if "annotation_proj" in request.data.keys():
            annotation_proj = request.data["annotation_proj"]
        else:
            annotation_proj = None
        if "ann_include_all" in request.data.keys():
            ann_include_all = True
        else:
            ann_include_all = False
        q = AbstractRelation.get_relation_class_of_name("".join(rel.split("-")))
        rel_match = re.match(r"([A-Z][a-z]+)([A-Z][a-z]+$)", rel)
        lst_nodes = []
        rel_a = "related_" + rel.split("-")[0]
        rel_b = "related_" + rel.split("-")[1]
        q_dict = {}
        q_list = None
        if rel_a == rel_b:
            rel_a += "A"
            rel_b += "B"
        if "search_source" in request.data.keys():
            source = request.data["search_source"]
            if source.startswith("cl:"):
                if q_list is not None:
                    q_list.append(
                        Q(**{rel_a + "__collection__id": int(source[3:])})
                        | Q(**{rel_b + "__collection__id": int(source[3:])})
                    )
                else:
                    q_dict[rel_a + "__collection__id"] = int(source[3:])
            else:
                if q_list is not None:
                    q_list.append(
                        Q(**{rel_a + "_id": int(source)})
                        | Q(**{rel_b + "_id": int(source)})
                    )
                else:
                    q_dict[rel_a + "_id"] = int(source)
        elif "search_source-autocomplete" in request.data.keys():
            source = request.data["search_source-autocomplete"]
            if q_list is not None:
                q_list.append(
                    Q(**{rel_a + "__name__icontains": source})
                    | Q(**{rel_b + "__name__icontains": source})
                )
            else:
                q_dict[rel_a + "__name__icontains"] = source
        if "search_target" in request.data.keys():
            target = request.data["search_target"]
            if target.startswith("cl:"):
                if q_list is not None:
                    q_list.append(
                        Q(**{rel_b + "__collection__id": int(target[3:])})
                        | Q(**{rel_a + "__collection__id": int(target[3:])})
                    )
                else:
                    q_dict[rel_b + "__collection__id"] = int(target[3:])
            else:
                if q_list is not None:
                    q_list.append(
                        Q(**{rel_a + "_id": int(target)})
                        | Q(**{rel_b + "_id": int(target)})
                    )
                else:
                    q_dict[rel_b + "_id"] = target
        elif "search_target-autocomplete" in request.data.keys():
            if len(request.data["search_target-autocomplete"]) > 0:
                target = request.data["search_target-autocomplete"]
                if q_list is not None:
                    q_list.append(
                        Q(**{rel_a + "__name__icontains": target})
                        | Q(**{rel_b + "__name__icontains": target})
                    )
                else:
                    q_dict[rel_b + "__name__icontains"] = target
        if kind:
            kind_ids = [int(kind.strip())]
            kind_ids_res = [int(kind.strip())]
            while len(kind_ids) > 0:
                id_a = kind_ids.pop()
                ids_b = VocabsBaseClass.objects.filter(
                    parent_class_id=id_a
                ).values_list("pk", flat=True)
                kind_ids.extend(ids_b)
                kind_ids_res.extend(ids_b)
            q_dict["relation_type_id__in"] = kind_ids_res
        if "start_date" in request.data.keys():
            if len(request.data["start_date"]) > 0:
                start_date_d = datetime.strptime(request.data["start_date"], "%d.%m.%Y")
                q_dict["end_date__gte"] = start_date_d
        if "end_date" in request.data.keys():
            if len(request.data["end_date"]) > 0:
                end_date_d = datetime.strptime(request.data["end_date"], "%d.%m.%Y")
                q_dict["start_date__lte"] = end_date_d
        if annotation_proj:
            annProj_filter = {"ann_proj": annotation_proj}
            if not ann_include_all:
                annProj_filter["include_all"] = False
            sr1 = (
                q.annotation_links.filter_ann_proj(**annProj_filter)
                .filter(**q_dict)
                .select_related(rel_a, rel_b)
            )
        else:
            sr1 = q.objects.filter(**q_dict)
        if q_list is not None:
            sr1 = sr1.filter(q_list[0])
        for node in sr1:
            if getattr(node, rel_a) not in lst_nodes:
                lst_nodes.append(getattr(node, rel_a))
            if getattr(node, rel_b) not in lst_nodes:
                lst_nodes.append(getattr(node, rel_b))
        sr1_1 = NetJsonEdgeSerializer(sr1, many=True)
        sr1_2 = NetJsonNodeSerializer(lst_nodes, many=True)
        res = {"nodes": sr1_2.data, "edges": sr1_1.data}
        return Response(res)


class SaveNetworkFiles(APIView):
    def post(self, request, format=None):
        file_name = request.data["file_name"]
        file_name_list = file_name.split(".")
        data = request.data["file"]
        nmb = False
        for file in os.listdir("downloads/"):
            if (
                fnmatch.fnmatch(file, file_name_list[0] + "_*." + file_name_list[1])
                or file == file_name
            ):
                if nmb:
                    nmb += 1
                else:
                    nmb = 1
        if nmb:
            file_name = file_name_list[0] + "_" + str(nmb) + "." + file_name_list[1]
        with open("downloads/" + file_name, "wb") as fw:
            data = data.replace('"edges"', '"links"', 1)
            fw.write(data.encode("utf-8"))
        res = {"test": True, "file_name": "/downloads/" + file_name}
        return Response(res)


class ResolveAbbreviations(APIView):
    parser_classes = (FileUploadParser,)

    def put(self, request, filename, format=None):
        file_obj = request.data["file"]
        txt = TextIOWrapper(file_obj, encoding="utf8")
        return Response(status=204)


class GetOrCreateEntity(APIView):
    def get(self, request):
        entity = request.query_params.get("entity2", None)
        uri = request.query_params.get("uri", None)
        if uri.startswith("http:"):
            ent = RDFParser(uri, entity.title()).get_or_create()
        else:
            r1 = re.search(r"^[^<]+", uri)
            r2 = re.search(r"<([^>]+)>", uri)
            q_d = dict()
            q_d["name"] = r1
            if r2:
                for x in r2.group(1).split(";"):
                    x2 = x.split("=")
                    q_d[x2[0].strip()] = x2[1].strip()
            if entity == "person":
                r1_2 = r1.group(0).split(",")
                if len(r1_2) == 2:
                    q_d["first_name"] = r1_2[1].strip()
                    q_d["name"] = r1_2[0].strip()
            ent = AbstractEntity.get_entity_class_of_name(entity).objects.create(**q_d)
        res = {
            "id": ent.pk,
            "url": reverse_lazy(
                "apis:apis_entities:generic_entities_edit_view",
                request=request,
                kwargs={"pk": ent.pk, "entity": entity},
            ),
        }
        return Response(res)


class GetRelatedPlaces(APIView):
    # map_qs = {
    #   'institution': Prefetch('personinstitution_set__related_institution__institutionplace_set', queryset=InstitutionPlace.objects.select_related('related_place'))
    # }

    def get(self, request):
        b_rel = PersonPlaceRelation.objects.get(
            name=getattr(settings, "BIRTH_REL_NAME", "")
        )
        d_rel = PersonPlaceRelation.objects.get(
            name=getattr(settings, "DEATH_REL_NAME", "")
        )
        person_pk = request.query_params.get("person_id", None)
        pers = Person.objects.get(pk=person_pk)
        relation_types = request.query_params.get("relation_types", None)
        place_pk = dict()
        res = []
        p = (
            PersonPlace.objects.select_related("related_place")
            .filter(related_person_id=person_pk, related_place__lat__isnull=False)
            .filter_for_user()
        )
        for pp in p:
            if pp.related_place_id not in place_pk:
                res.append(
                    (
                        pp.related_place,
                        [
                            (
                                pp.relation_type,
                                None,
                                pers.start_date
                                if b_rel.pk == pp.relation_type_id
                                else pp.start_date,
                                pers.end_date
                                if d_rel.pk == pp.relation_type_id
                                else pp.end_date,
                                pp,
                            )
                        ],
                    )
                )
                place_pk[pp.related_place_id] = len(res) - 1
            else:
                res[place_pk[pp.related_place_id]][1].append(
                    (
                        pp.relation_type,
                        None,
                        pers.start_date
                        if b_rel.pk == pp.relation_type_id
                        else pp.start_date,
                        pers.end_date
                        if d_rel.pk == pp.relation_type_id
                        else pp.end_date,
                        pp,
                    )
                )
        for pi in PersonInstitution.objects.filter(
            related_person_id=person_pk
        ).filter_for_user():
            inst = pi.related_institution
            rel_type = getattr(
                settings,
                "APIS_LOCATED_IN_ATTR",
                [
                    "situated in",
                ],
            )
            ipl_rel = InstitutionPlaceRelation.objects.filter(
                name__in=rel_type
            ).values_list("pk", flat=True)
            plc = InstitutionPlace.objects.filter(
                relation_type_id__in=ipl_rel, related_institution=inst
            )
            if plc.count() == 1:
                if plc[0].related_place_id not in place_pk:
                    res.append(
                        (
                            plc[0].related_place,
                            [
                                (
                                    pi.relation_type,
                                    pi.related_institution,
                                    pi.start_date,
                                    pi.end_date,
                                    pi,
                                )
                            ],
                        )
                    )
                    place_pk[plc[0].related_place_id] = len(res) - 1
                else:
                    res[place_pk[plc[0].related_place_id]][1].append(
                        (
                            pi.relation_type,
                            pi.related_institution,
                            pi.start_date,
                            pi.end_date,
                            pi,
                        )
                    )
        res = GeoJsonSerializerTheme(res, many=True)
        return Response(res.data)


class LifePathViewset(APIView):
    def get(self, request, pk):
        b_rel = PersonPlaceRelation.objects.get(
            name=getattr(settings, "BIRTH_REL_NAME", "")
        )
        d_rel = PersonPlaceRelation.objects.get(
            name=getattr(settings, "DEATH_REL_NAME", "")
        )
        pb_pd = [b_rel.pk, d_rel.pk]
        lst_inst = list(
            PersonInstitution.objects.filter(
                Q(related_person_id=pk),
                Q(start_date__isnull=False) | Q(end_date__isnull=False),
            ).filter_for_user()
        )
        lst_place = list(
            PersonPlace.objects.filter(
                Q(related_person_id=pk),
                Q(start_date__isnull=False)
                | Q(end_date__isnull=False)
                | Q(relation_type_id__in=pb_pd),
            ).filter_for_user()
        )
        comb_lst = lst_inst + lst_place
        p1 = Person.objects.get(pk=pk)
        if p1.start_date:
            for e in comb_lst:
                if e.relation_type_id == b_rel.pk:
                    e.start_date = p1.start_date
        if p1.end_date:
            for e in comb_lst:
                if e.relation_type_id == d_rel.pk:
                    e.start_date = p1.end_date
        data = LifePathSerializer(comb_lst, many=True).data
        data = [d for d in data if d is not None]
        data = sorted(data, key=lambda i: i["year"])

        return Response(data)



import django_tables2 as tables
from django.db.models import F
from django.utils.html import format_html

from .models import Uri


# generic order_FOO methods for start_date_written and end_date_written to be used in all kinds of tables where a queryset is loaded
# whose model has start_date_written and end_date_written. These methods order the *_date_written by their corresponding
# parsed date objects.
#
# In order to use them in a table using django-tables2, the table class must be assigned *_date_written columns
# (either through fields in Meta inner class or through custom Column() definitions).
# Important however is that the names between the columns and the order methods are correlated to each other.
#
# The django-tables2 plugin searches for each column through the table class for corresponding order_* and render_* methods
# E.g. to define custom ordering and rendering for a field column 'name' the table class must have
# 'order_name' and 'render_name' methods. This approach is called order_FOO and render_FOO by django-tables2.
#
# So in order to use these functions here, define within the table class
# 1.) the respective column
# 2.) a correspondingly named method for each column
# 3.) assign to this method one of the two functions below
#
# In the case of e.g. start_date_written,
# the table class thus must have a method / class variable 'order_start_date_written = generic_order_start_date_written'

def generic_order_start_date_written(self, queryset, is_descending):
    if is_descending:
        queryset = queryset.order_by(F("start_date").desc(nulls_last=True)) #nulls_last=True puts null values behind non-null values
    else:
        queryset = queryset.order_by(F("start_date").asc(nulls_last=True))

    return (queryset, True)


def generic_order_end_date_written(self, queryset, is_descending):
    if is_descending:
        queryset = queryset.order_by(F("end_date").desc(nulls_last=True))
    else:
        queryset = queryset.order_by(F("end_date").asc(nulls_last=True))

    return (queryset, True)



def helper_render_date(value, var_date, var_start_date, var_end_date):
    """
    helper function to avoid duplicated code. It checks the various sub-dates of a model's date field for them being None
    or having values. If a field is None then check for the next, and if all are None, return '—' inditcating no value.

    If there are values, use them as mouse overlay helptext to inform the user about the parsing result behind a written
    date field.

    :param value: str : the *_date_written (either start_date_written or end_date_written) field of an entity or relation
    :param var_date: datetime : either the precisely parsed date or the average in between two dates when *_date_written is a range
    :param var_start_date: datetime : The sub-date of var_date, indicating the start date of the range
    :param var_end_date: datetime : The sub-date of var_date, indicating the end date of the range
    :return: html string : which has the value of the written date and the parsed dates as mouse overlay helptext
    """


    # Various if-else branches checking which of the date fields are not None and should be used

    if var_start_date != None and var_end_date != None:

        overlay_help_text = str(var_start_date) + " - " + str(var_end_date)

    elif var_date != None:

        overlay_help_text = str(var_date)

    else:

        return "—"

    return format_html("<abbr title='" + overlay_help_text + "'>" + value + "</b>")


# Again this function serves a generic purpose and must be assigned as class method to django-tables2 tables.Table class
# The whole logic is very similare to the generic_order_* functions above, so see their comments for more details.
def generic_render_start_date_written(self, record, value):

    return helper_render_date(
        value = value,
        var_date = record.start_date,
        var_start_date = record.start_start_date,
        var_end_date = record.start_end_date
    )


def generic_render_end_date_written(self, record, value):

    return helper_render_date(
        value = value,
        var_date = record.end_date,
        var_start_date = record.end_start_date,
        var_end_date = record.end_end_date
    )




class UriTable(tables.Table):
    id = tables.LinkColumn()
    entity = tables.TemplateColumn(
        "<a href='{{ record.entity.get_absolute_url }}'>{{ record.entity }}</a>",
        orderable=True, verbose_name="related Entity"
    )
    ent_type = tables.TemplateColumn(
        "{{ record.entity.get_child_class }}",
        orderable=False, verbose_name="Entity Type"
    )

    class Meta:
        model = Uri
        sequence = ('id', 'uri')
        attrs = {"class": "table table-responsive table-hover"}
